"""Handle pairtree (n=2) and similar directory structures.

Note that saying Pairtree for a given n, default 2, is not sufficient
to define the object layout. The Pairtree specification
https://confluence.ucop.edu/display/Curation/PairTree
encourages the use of object encapsulation (section 3) but does
not prescribe a particular method. In this implementation the default
it to encapsulate in a directory with the complete encoded identifier
name.

Makes use of encoding and decoding functions from Ben O'Steen's
implementation in the pairtree module
(https://github.com/benosteen/pairtree).
"""
import os
import os.path
from pairtree import id_encode, id_decode

from .dispositor import Dispositor


class Ntree(Dispositor):
    """Class to support pairtree and related dispositions."""

    def __init__(self, n=2, encapsulate=True):
        """Initialize Dispositor."""
        super(Ntree, self).__init__()
        self.n = n
        self.encapsulate = encapsulate

    def encode(self, identifier):
        """Pairtree encode identifier."""
        return id_encode(identifier)

    def decode(self, identifier):
        """Pairtree decode identifier."""
        return id_decode(identifier)

    def identifier_to_path(self, identifier):
        """Convert identifier to path relative to root."""
        identifier = self.encode(identifier)
        path = ''
        id_remains = identifier
        segments = []
        while len(id_remains) > self.n:
            segments.append(id_remains[0:self.n])
            id_remains = id_remains[self.n:]
        segments.append(id_remains)
        if self.encapsulate:
            segments.append(identifier)
        return os.path.join(*segments)

    def relative_path_to_identifier(self, path):
        """Convert relative path to identifier."""
        if self.encapsulate:
            (path, encap_id) = os.path.split(path)
        # Combine all directories
        identifier = ''.join(path.split(os.sep))
        if self.encapsulate and identifier != encap_id:
            raise Exception("Bad ntree path: id from path (%s) does not match encapsulation id (%s)" % (identifier, encap_id))
        return identifier



"""OCFL Object Store library."""
import hashlib
import json
import os
import os.path
import re
import logging
from shutil import copyfile, copytree
try:
    from urllib.parse import quote_plus  # py3
except ImportError:                      # pragma: no cover -- py2
    from urllib import quote_plus        # pragma: no cover -- py2

from .digest import file_digest
from .disposition import get_dispositor
from .namaste import find_namastes, Namaste
from .object import Object
from .validator import Validator


class StoreException(Exception):
    """Exception class for OCFL Store."""

    pass


class Store(object):
    """Class for handling OCFL Object Stores."""

    def __init__(self, root=None, disposition=None, lax_digests=False):
        """Initialize OCFL Object Store."""
        self.root = root
        self.disposition = disposition
        self.lax_digests = lax_digests
        self._dispositor = None
        #
        self.declaration_tvalue = 'ocfl_1.0'
        self.num_traversal_errors = 0

    @property
    def spec_file(self):
        """Path of optional storage root spec file."""
        return os.path.join(self.root, 'ocfl_1.0.txt')

    @property
    def layout_file(self):
        """Path of optional storage root layout file."""
        return os.path.join(self.root, 'ocfl_layout.json')

    @property
    def dispositor(self):
        """Instance of dispositor class.

        Lazily initialized.
        """
        if not self._dispositor:
            self._dispositor = get_dispositor(disposition=self.disposition)
        return self._dispositor

    def traversal_error(self, message):
        """Record error traversing storage root."""
        self.num_traversal_errors += 1
        logging.error(message)

    def object_path(self, identifier):
        """Path to OCFL object with given identifier."""
        return os.path.join(self.root, self.dispositor.identifier_to_path(identifier))

    def initialize(self):
        """Initialize an object store."""
        if os.path.exists(self.root):
            raise StoreException("OCFL Object Store root %s already exists, aborting!" % (self.root))
        os.makedirs(self.root)
        logging.debug("Created root directory %s" % (self.root))
        # Create root declaration
        Namaste(d=0, content=self.declaration_tvalue).write(self.root)
        # Create a layout declaration
        if self.disposition is not None:
            with open(self.layout_file, 'w') as fh:
                layout = {'key': self.disposition,
                          'description': "Non-standard layout from ocfl-py disposition -- FIXME"}
                json.dump(layout, fh, sort_keys=True, indent=2)
        logging.info("Created OCFL storage root %s" % (self.root))

    def check_root_structure(self):
        """Check the storage root.

        Raises StoreException if there is an error.
        """
        if not os.path.exists(self.root):
            raise StoreException("Storage root %s does not exist!" % (self.root))
        if not os.path.isdir(self.root):
            raise StoreException("Storage root %s is not a directory" % (self.root))
        # Storage root declaration
        namastes = find_namastes(0, self.root)
        if len(namastes) == 0:
            raise StoreException("Storage root %s lacks required 0= declaration file" % (self.root))
        elif len(namastes) > 1:
            raise StoreException("Storage root %s has more than one 0= style declaration file" % (self.root))
        elif namastes[0].tvalue != self.declaration_tvalue:
            raise StoreException("Storage root %s declaration file not as expected, got %s" % (self.root, namastes[0].filename))
        elif not namastes[0].content_ok(self.root):
            raise StoreException("Storage root %s required declaration file %s has invalid content" % (self.root, namastes[0].filename))
        # Specification file and layout file
        if os.path.exists(self.spec_file) and not os.path.isfile(self.spec_file):
            raise StoreException("Storage root %s includes a specification entry that isn't a file" % (self.root))
        if os.path.exists(self.layout_file):
            self.parse_layout_file()
        # Other files are allowed...
        return True

    def parse_layout_file(self):
        """Read and parse layout file.

        Returns key and description on success, otherwise raises a StorageException.
        """
        try:
            with open(self.layout_file) as fh:
                layout = json.load(fh)
            if type(layout) != dict:
                raise StoreException("Storage root %s has layout file that isn't a JSON object" % (self.root))
            elif 'key' not in layout or 'description' not in layout:
                raise StoreException("Storage root %s has layout file doesn't have required key and description entries" % (self.root))
            return layout['key'], layout['description']
        except Exception as e:  # FIXME - more specific?
            raise StorageException("Sorage root %s has layout file can't be read (%s)" % (self.root, str(e)))

    def object_paths(self):
        """Generator for object paths for every obects in the store.

        Yields (dirpath) that is the path to the directory for each object
        located.

        Will log any errors seen while traversing the directory tree under the
        storage root.
        """
        for (dirpath, dirnames, filenames) in os.walk(self.root, followlinks=True):
            # Ignore files in root
            if dirpath != self.root:
                if (len(dirnames) + len(filenames)) == 0:
                    self.traversal_error("Empty directory %d" % (dirpath))
                # Is this directory an OCFL object (or root?). Look for any 0= file.
                zero_eqs = list(filter(lambda x: x.startswith('0='), filenames))
                if len(zero_eqs) > 1:
                    self.traversal_error("Multiple 0= declaration files in %s, ignoring and not descending" % (dirpath))
                    dirnames = []
                elif len(zero_eqs) == 1:
                    declaration = zero_eqs[0]
                    match = re.match(r'''0=ocfl_object_(\d+\.\d+)''', declaration)
                    if match and match.group(1) == '1.0':
                        yield (dirpath)
                    elif match:
                        self.traversal_error("Object with unknown version %s in %s, ignoring" % (match.group(1), dirpath))
                    else:
                        self.traversal_error("Object with unrecognized declaration %s in %s, ignoring" % (declaration, dirpath))
                    # don't descend
                    dirnames = []

    def list(self):
        """List contents of storage."""
        self.check_root_structure()
        num_objects = 0
        for dirpath in self.object_paths():
            # Parse inventory to extract id
            id = Object().id_from_inventory(dirpath)
            print("%s -- id=%s" % (os.path.relpath(dirpath, self.root), id))
            num_objects += 1
        # FIXME - do some stuff in here
        logging.info("Found %d OCFL Objects under root %s" % (num_objects, self.root))

    def validate(self, validate_objects=True, show_warnings=False, show_errors=True, check_digests=True):
        """Validate storage root and optionally all objects."""
        valid = True
        try:
            self.check_root_structure()
            logging.info("Storage root structure is VALID")
        except StoreException as e:
            valid = False
            logging.info("Storage root structure is INVALID (%s)" % (str(e)))
        num_objects = 0
        good_objects = 0
        for dirpath in self.object_paths():
            if validate_objects:
                validator = Validator(check_digests=check_digests,
                                      lax_digests=self.lax_digests,
                                      show_warnings=show_warnings)
                if validator.validate(dirpath):
                    good_objects += 1
                else:
                    logging.info("Object at %s in INVALID" % (dirpath))
                messages = validator.__str__(prefix='[[' + dirpath + ']]')  # FIXME - how to show warnings sensibly?
                if messages != '':
                    print(messages)
                num_objects += 1
        if validate_objects:
            if good_objects == num_objects:
                logging.info("Objects checked: %d / %d are VALID" % (good_objects, num_objects))
            else:
                valid = False
                logging.info("Objects checked: %d / %d are INVALID" % (num_objects - good_objects, num_objects))
        else:
            logging.info("Not checking OCFL objects")
        if self.num_traversal_errors > 0:
            valid = False
            logging.info("Encountered %d errors traversing storage root" % (self.num_traveral_errors))
        # FIXME - do some stuff in here
        if valid:
            logging.info("Storage root %s is VALID", self.root)
        else:
            logging.info("Storage root %s is INVALID", self.root)
        return valid

    def add(self, object_path):
        """Add pre-constructed object from object_path."""
        # Sanity check
        o = Object()
        inventory = o.parse_inventory(object_path)
        identifier = inventory['id']
        path = self.object_path(identifier)
        logging.info("Copying from %s to %s" % (object_path, path))
        try:
            copytree(object_path, path)
            logging.info("Copied")
        except Exception as e:
            logging.error("Copy failed: " + str(e))
            raise StoreException("Add object failed!")



#!/usr/bin/env python
"""Validate an OCFL Object."""
import argparse
import logging
import ocfl
import sys

parser = argparse.ArgumentParser(
    description='Validate one or more OCFL objects or storage roots. By default '
    'shows any errors or warnings, and final validation status. Use -q to show '
    'only errors, -Q to show only validation status.')
parser.add_argument('path', type=str, nargs='*',
                    help='OCFL object or storage root path(s) to validate')
parser.add_argument('--quiet', '-q', action='store_true',
                    help="Be quiet, do not show warnings")
parser.add_argument('--very-quiet', '-Q', action='store_true',
                    help="Be very quiet, show only validation status (implies -q)")
parser.add_argument('--lax-digests', action='store_true',
                    help='Allow use of any known digest')
parser.add_argument('--no-check-digests', action='store_true',
                    help='Do not check digest values')

ocfl.add_shared_args(parser)
args = parser.parse_args()
ocfl.check_shared_args(args)

logging.basicConfig(level=logging.INFO if args.verbose else logging.WARN)
log = logging.getLogger(name="ocfl-validate")
log.setLevel(level=logging.INFO if args.verbose else logging.WARN)

if len(args.path) == 0:
    print("No OCFL paths specified, nothing to do! (Use -h for help)")

num = 0
num_good = 0
num_paths = len(args.path)
for path in args.path:
    num += 1
    path_type = ocfl.find_path_type(path)
    if path_type == 'object':
        log.info("Validating OCFL Object at " + path)
        obj = ocfl.Object(lax_digests=args.lax_digests)
        if obj.validate(path,
                        show_warnings=not args.quiet and not args.very_quiet,
                        show_errors=not args.very_quiet,
                        check_digests=not args.no_check_digests):
            num_good += 1
    elif path_type == 'root':

        log.info("Validating OCFL Storage Root at " + path)
        store = ocfl.Store(root=path,
                           lax_digests=args.lax_digests)
        if store.validate(show_warnings=not args.quiet and not args.very_quiet,
                          show_errors=not args.very_quiet,
                          check_digests=not args.no_check_digests):
            num_good += 1
    else:
        log.error("Bad path %s (%s)" % (path, path_type))
    if num_paths > 1:
        log.info(" [%d / %d paths validated, %d / %d VALID]\n" % (num, num_paths, num_good, num))
if num_good != num:
    sys.exit(1)



"""Digest handling for OCFL."""
import hashlib


def _file_digest(filename, digester):
    """Generate a digest for filename using the supplied digester object.

    Like haslib.sha256 and hashlib.sha512, the digester object must
    support the .update() and .hexdigest() methods.
    """
    BUFSIZE = 64 * 1024  # 64kB for want of better info...
    with open(filename, 'rb', buffering=0) as f:
        for b in iter(lambda: f.read(BUFSIZE), b''):
            digester.update(b)
    return digester.hexdigest()


def file_digest(filename, digest_type='sha512'):
    """Digest of digest_type for file filename in normalized form.

    Supports digest_type values from OCFL spec:
        'md5', 'sha1', 'sha256', 'sha512', 'blake2b-512'
    and from OCFL extensions
        'blake2b-160', 'blake2b-256', 'blake2b-384'
    and a truncated sha512 and sha256 useful for examples
        'sha512-spec-ex',  'sha256-spec-ex'

    Raises an exception if the digest_type is not supported.
    """
    # From spec
    if digest_type == 'sha512':
        return _file_digest(filename, hashlib.sha512())
    elif digest_type == 'sha256':
        return _file_digest(filename, hashlib.sha256())
    elif digest_type == 'sha1':
        return _file_digest(filename, hashlib.sha1())
    elif digest_type == 'md5':
        return _file_digest(filename, hashlib.md5())
    elif digest_type == 'blake2b-512':
        return _file_digest(filename, hashlib.blake2b())
    # From extensions
    elif digest_type == 'blake2b-160':
        return _file_digest(filename, hashlib.blake2b(digest_size=20))
    elif digest_type == 'blake2b-256':
        return _file_digest(filename, hashlib.blake2b(digest_size=32))
    elif digest_type == 'blake2b-384':
        return _file_digest(filename, hashlib.blake2b(digest_size=48))
    # Specification examples: 15/6 chars ... 3 chars. The truncated
    # sha512 is twice as many chars as the truncated sha256 to give
    # a appropriate impression in examples
    elif digest_type == 'sha512-spec-ex':
        d = _file_digest(filename, hashlib.sha512())
        return d[:15] + '...' + d[-3:]
    elif digest_type == 'sha256-spec-ex':
        d = _file_digest(filename, hashlib.sha256())
        return d[:6] + '...' + d[-3:]
    else:
        raise ValueError("Unsupport digest type %s" % (digest_type))


DIGEST_REGEXES = {
    'sha512': r'''^[0-9a-fA-F]{128}$''',
    'sha256': r'''^[0-9a-fA-F]{64}$''',
    'sha1': r'''^[0-9a-fA-F]{40}$''',
    'md5': r'''^[0-9a-fA-F]{32}$''',
    'blake2b-512': r'''^[0-9a-fA-F]{128}$''',
    'blake2b-384': r'''^[0-9a-fA-F]{96}$''',
    'blake2b-256': r'''^[0-9a-fA-F]{64}$''',
    'blake2b-160': r'''^[0-9a-fA-F]{40}$''',
    'sha512-spec-ex': r'''^[0-9a-f]{15}\.\.\.[0-9a-f]{3}$''',
    'sha256-spec-ex': r'''^[0-9a-f]{6}\.\.\.[0-9a-f]{3}$'''
}


def digest_regex(digest_type='sha512'):
    """Regex to be used to check the un-normalized form of a digest string."""
    try:
        return DIGEST_REGEXES[digest_type]
    except KeyError:
        raise ValueError("Unsupport digest type %s" % (digest_type))


def normalized_digest(digest, digest_type='sha512'):
    """Normalized version of the digest that enables string comparison.

    All forms (except the spec example forms) are case insensitive. We
    use lowercase as the normalized form.
    """
    if digest_type != 'sha512-spec-ex' and digest_type != 'sha256-spec-ex':
        return digest.lower()
    return digest



#!/usr/bin/env python
"""OCFL Storage Root Tool."""
import argparse
import logging
import ocfl
import sys

parser = argparse.ArgumentParser(description='Manpulate or validate an OCFL Storage Root.',
                                 formatter_class=argparse.ArgumentDefaultsHelpFormatter)
parser.add_argument('--root', required=True,
                    help='OCFL Storage Root (must be supplied)')
parser.add_argument('--disposition', '-d', default=None,
                    help='Disposition of objects under root')
parser.add_argument('--quiet', '-q', action='store_true',
                    help="be quiet, do not show warnings")

commands = parser.add_mutually_exclusive_group(required=True)
# These commands are actions on the store
commands.add_argument('--init', action='store_true',
                      help='Initialize an object store at specified --root')
commands.add_argument('--list', action='store_true',
                      help='List contents of an object store at --root')
commands.add_argument('--validate', action='store_true',
                      help='Validate an object store at --root and its contents')
commands.add_argument('--add', action='store_true',
                      help='Add object at --src to object store at --root')
commands.add_argument('--purge', action='store_true',
                      help='Purge (delete) an object --id from object store at --root')
# and these commands act on an object in the store
commands.add_argument('--create-object', action='store_true',
                      help='Create an new object with version 1 from objdir')
commands.add_argument('--show-object', action='store_true',
                      help='Show versions and files in an OCFL object')
commands.add_argument('--validate-object', action='store_true',
                      help='Validate an OCFL object')

# Object property settings
parser.add_argument('--id', default=None,
                    help='identifier of object')
parser.add_argument('--src', default=None,
                    help='source path of object or version')
parser.add_argument('--digest', default='sha512',
                    help='digest type to use')
parser.add_argument('--fixity', action='append',
                    help='add fixity type to add')

# Version metadata and object settings
ocfl.add_version_metadata_args(parser)
ocfl.add_object_args(parser)
ocfl.add_shared_args(parser)
args = parser.parse_args()
ocfl.check_shared_args(args)

logging.basicConfig(level=logging.INFO if args.verbose else logging.WARN)

try:
    store = ocfl.Store(root=args.root,
                       disposition=args.disposition,
                       lax_digests=args.lax_digests)
    if args.init:
        store.initialize()
    elif args.list:
        store.list()
    elif args.validate:
        store.validate(show_warnings=not args.quiet)
    elif args.add:
        if not args.src:
            raise ocfl.StoreException("Must specify object path with --src")
        store.add(object_path=args.src)
    elif args.purge:
        logging.error("purge not implemented")
    elif args.show_object or args.validate_object:
        if not args.id:
            raise ocfl.StoreException("Must specify id to act on an object in the store")
        objdir = store.object_path(args.id)
        obj = ocfl.Object(identifier=args.id,
                          digest_algorithm=args.digest,
                          forward_delta=not args.no_forward_delta,
                          dedupe=not args.no_dedupe,
                          skips=args.skip,
                          ocfl_version=args.ocfl_version,
                          fixity=args.fixity)
        if args.show_object:
            obj.show(objdir)
        else:
            logging.error("create/build/validate not implemented")
    else:
        logging.warn("Nuttin' happenin' 'round 'ere.")
except (ocfl.StoreException, ocfl.ObjectException) as e:
    logging.error(str(e))
    sys.exit(1)



"""Identity mapping of identifier to directory structure."""
import os

from .dispositor import Dispositor


class Identity(Dispositor):
    """Class to support trivial identity disposition."""

    def __init__(self):
        """Initialize Dispositor."""
        super(Identity, self).__init__()

    def identifier_to_path(self, identifier):
        """Convert identifier to path relative to root."""
        return self.encode(identifier)

    def relative_path_to_identifier(self, path):
        """Convert relative path to identifier.

        It is an error to include more than one path segment so raise
        and exception if os.sep exists in the path.
        """
        if os.sep in path:
            raise Exception("Relative path in Identity dispositor must not have multiple path segments!")
        return self.decode(path)



#!/usr/bin/env python
"""OCFL Object and Inventory Builder."""
import argparse
import logging
import ocfl
import os.path
import sys


class FatalError(Exception):
    """Exception class for conditions that should abort with message."""

    pass


def parse_arguments():
    """Parse command line arguments."""
    parser = argparse.ArgumentParser(description='Manipulate or validate an OCFL object or inventory.',
                                     formatter_class=argparse.ArgumentDefaultsHelpFormatter)

    # This should really be done with add_mutually_exclusive_group() but
    # the current argparse doesn't support grouping and a title for that
    # so use a plain group instead and will check later
    commands = parser.add_argument_group(title="Commands (exactly one required)")
    commands.add_argument('--create', action='store_true',
                          help='Create a new object with version 1 files in --srcdir or from --srcbag')
    commands.add_argument('--build', action='store_true',
                          help='Build a new object from version directories in --srcdir')
    commands.add_argument('--update', action='store_true',
                          help='Update an object by adding a new version from files in --srcdir or from --srcbag')
    commands.add_argument('--show', action='store_true',
                          help='Show versions and files in an OCFL object')
    commands.add_argument('--validate', action='store_true',
                          help='Validate an OCFL object (use ocfl-validate.py for more control)')
    commands.add_argument('--extract', action='store', default=None,
                          help='Extract a specific version (or "head") into --dstdir')

    src_params = parser.add_argument_group(title="Source files")
    src_params.add_argument('--srcdir', '--src', action='store',
                            help='Source directory path')
    src_params.add_argument('--srcbag', action='store',
                            help='Source Bagit bag path (alternative to --srcdir)')

    obj_params = parser.add_argument_group(title="OCFL object parameters")
    obj_params.add_argument('--digest', default='sha512',
                            help='Digest algorithm to use')
    obj_params.add_argument('--fixity', action='append',
                            help='Add fixity type to add')
    obj_params.add_argument('--id', default=None,
                            help='Identifier of object')
    obj_params.add_argument('--dstdir', '--dst', action='store', default='/tmp/ocfl-out',
                            help='Destination directory path')
    obj_params.add_argument('--dstbag', action='store',
                            help='Destination Bagit bag path (alternative to --dstdir)')

    # Version metadata and object settings
    ocfl.add_version_metadata_args(obj_params)
    ocfl.add_object_args(obj_params)
    ocfl.add_shared_args(parser)
    args = parser.parse_args()
    ocfl.check_shared_args(args)

    # Require command and only one command
    cmds = ['create', 'build', 'update', 'show', 'validate', 'extract']
    num_cmds = 0
    for cmd in cmds:
        if getattr(args, cmd):
            num_cmds += 1
    if num_cmds != 1:
        raise FatalError("Exactly one command (%s) must be specified" % ', '.join(cmds))

    # Must not specify both srcdir and srcbag
    if args.srcdir and args.srcbag:
        raise FatalError("Must not specify both --srcdir and --srcbag")

    return args


def do_object_operation(args):
    """Implement object operations in a way that can be reused by ocfl-store.py."""
    obj = ocfl.Object(id=args.id,
                      digest_algorithm=args.digest,
                      filepath_normalization=args.normalization,
                      skips=args.skip,
                      forward_delta=not args.no_forward_delta,
                      dedupe=not args.no_dedupe,
                      lax_digests=args.lax_digests,
                      ocfl_version=args.ocfl_version,
                      fixity=args.fixity)
    if args.create:
        srcdir = args.srcdir
        metadata = ocfl.VersionMetadata(args)
        if args.srcbag is not None:
            srcdir = ocfl.bag_as_source(args.srcbag, metadata)
            if metadata.id is not None:
                if obj.id:
                    if obj.id != metadata.id:
                        raise FatalError("Identifier specified (%s) and identifier from Bagit bag (%s) do not match!" % (obj.id, metadata.id))
                else:
                    obj.id = metadata.id
        elif args.srcdir is None:
            raise FatalError("Must specify either --srcdir or --srcbag containing v1 files when creating an OCFL object!")
        obj.create(srcdir=srcdir,
                   metadata=metadata,
                   objdir=args.objdir)
    elif args.build:
        if args.srcdir is None:
            raise FatalError("Must specify --srcdir containing version directories when building an OCFL object!")
        metadata = ocfl.VersionMetadata(args)
        obj.build(srcdir=args.srcdir,
                  metadata=metadata,
                  objdir=args.objdir)
    elif args.update:
        srcdir = args.srcdir
        metadata = ocfl.VersionMetadata(args)
        if args.srcbag is not None:
            srcdir = ocfl.bag_as_source(args.srcbag, metadata)
        elif args.srcdir is None:
            raise FatalError("Must specify either --srcdir or --srcbag containing new version files when updating an OCFL object!")
        obj.update(objdir=args.objdir,
                   srcdir=srcdir,
                   metadata=metadata)
    elif args.show:
        obj.show(objdir=args.objdir)
    elif args.validate:
        obj.validate(objdir=args.objdir)
    elif args.extract:
        if args.dstdir and args.dstbag:
            args.dstdir = None  # Override dstdir if dstbag specified
        version = args.extract
        dst = os.path.join(args.dstdir or args.dstbag)
        metadata = obj.extract(objdir=args.objdir,
                               version=version,
                               dstdir=dst)
        if args.dstdir:
            print("Extracted content for %s in %s" % (version, dst))
        else:  # args.dstbag
            ocfl.bag_extracted_version(dst, metadata)
            print("Extracted content for %s saved as Bagit bag in %s" % (version, dst))
    else:
        raise FatalError("Command argument not supported!")


if __name__ == "__main__":
    try:
        args = parse_arguments()
        logging.basicConfig(level=logging.INFO if args.verbose else logging.WARN)
        do_object_operation(args)
    except (FatalError, ocfl.ObjectException) as e:
        # Show message but otherwise exit quietly
        print('Error - ' + str(e))
        sys.exit(1)



#!/usr/bin/env python
"""OCFL inventory sidecar generator and updater."""
import argparse
import logging
import ocfl
import os.path

INVENTORY_NAME = "inventory.json"


def parse_arguments():
    """Parse command line arguments."""
    parser = argparse.ArgumentParser(description="Update OCFL inventory sidecar file",
                                     formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument("path", type=str, nargs="*",
                        help="OCFL inventory files or directories containing them")
    parser.add_argument("--digest", default=None,
                        help="Digest algorithm to use overriding any in inventory")
    ocfl.add_shared_args(parser)
    args = parser.parse_args()
    ocfl.check_shared_args(args)
    return args


def create_sidecar(dir):
    """Create sidecar for inventory in dir."""
    inventory_path = os.path.join(dir, INVENTORY_NAME)
    if not os.path.isfile(inventory_path):
        logging.error("Ignoring path %s because there is not file %s." % (dir, inventory_path))
    else:
        object = ocfl.Object()
        if args.digest is not None:
            object.digest_algorithm = args.digest
        else:  # Read inventory in the hope of setting digest_algoritm
            try:
                object.parse_inventory(dir)
            except ocfl.ObjectException as e:
                logging.warning("Failed to read inventory in directory %s (%s)" % (dir, str(e)))
        sidecar = object.write_inventory_sidecar(dir)
        logging.info("Written sidecar file %s" % (sidecar))


if __name__ == "__main__":
    args = parse_arguments()
    logging.basicConfig(level=logging.INFO if args.verbose else logging.WARN)
    paths = ["."] if len(args.path) == 0 else args.path
    for path in paths:
        logging.info("Looking at path %s" % (path))
        if os.path.isdir(path):
            create_sidecar(path)
        else:
            (dir, filename) = os.path.split(path)
            if filename == INVENTORY_NAME:
                create_sidecar(dir)
            else:
                logging.error("Ignoring path %s with filename that is not inventory.json")
    print("Done.")



"""OCFL Validator.

Philosophy of this code is to keep it separate from the implementations
of Store, Object and Version used to build and manipulate OCFL data, but
to leverage lower level functions such as digest creation etc.. Code style
is plain/verbose with detailed and specific validation errors that might
help someone debug an implementation.
"""
import json
import os
import os.path
import re

from .digest import file_digest, normalized_digest
from .inventory_validator import InventoryValidator
from .namaste import find_namastes, NamasteException
from .validation_logger import ValidationLogger
from .w3c_datetime import str_to_datetime


class ValidatorAbortException(Exception):
    """Exception class to bail out of validation."""

    pass


class Validator(object):
    """Class for OCFL Validator."""

    def __init__(self, log=None, show_warnings=False, show_errors=True, check_digests=True, lax_digests=False, lang='en'):
        """Initialize OCFL validator."""
        self.log = log
        self.check_digests = check_digests
        self.lax_digests = lax_digests
        if self.log is None:
            self.log = ValidationLogger(show_warnings=show_warnings, show_errors=show_errors, lang=lang)
        self.registered_extensions = ['FIXME']  # FIXME - add names when something registered
        self.initialize()

    def initialize(self):
        """Initialize object state.

        Must be called between attempts to validate objects.
        """
        self.digest_algorithm = 'sha512'
        self.content_directory = 'content'
        self.inventory_digest_files = {}  # index by version_dir, algorithms may differ
        self.root_inv_validator = None

    def __str__(self, prefix=''):
        """String representation of validation log."""
        return self.log.__str__(prefix=prefix)

    def validate(self, path):
        """Validate OCFL object at path.

        Returns True if valid (warnings permitted), False otherwise.
        """
        self.initialize()
        if not os.path.isdir(path):
            self.log.error('E003c', path=path)
            return False
        # Object declaration
        namastes = find_namastes(0, path)
        if len(namastes) == 0:
            self.log.error('E003a')
        elif len(namastes) > 1:
            self.log.error('E003b', files=len(namastes))
        elif not namastes[0].content_ok(path):
            self.log.error('E007')
        # Object root inventory file
        inv_file = os.path.join(path, 'inventory.json')
        if not os.path.exists(inv_file):
            self.log.error('E034')
            return False
        try:
            inventory, inv_validator = self.validate_inventory(inv_file)
            self.root_inv_validator = inv_validator
            all_versions = inv_validator.all_versions
            self.content_directory = inv_validator.content_directory
            self.digest_algorithm = inv_validator.digest_algorithm
            self.validate_inventory_digest(inv_file, self.digest_algorithm)
            if self.log.num_errors > 0:
                # Don't look at storage if inventory fails validation
                return False
            # Object root
            self.validate_object_root(path, all_versions)
            # Version inventory files
            self.validate_version_inventories(path, inventory, all_versions)
            # Object content
            self.validate_content(path, inventory, all_versions)
        except ValidatorAbortException:
            pass
        return self.log.num_errors == 0

    def validate_inventory(self, inv_file, where='root'):
        """Validate a given inventory file, record errors with self.log.error().

        Returns inventory object for use in later validation
        of object content. Does not look at anything else in the
        object itself.
        """
        try:
            with open(inv_file) as fh:
                inventory = json.load(fh)
        except json.decoder.JSONDecodeError as e:
            self.log.error('E033', where=where, explanation=str(e))
            raise ValidatorAbortException
        inv_validator = InventoryValidator(log=self.log, where=where,
                                           lax_digests=self.lax_digests)
        inv_validator.validate(inventory)
        return inventory, inv_validator

    def validate_inventory_digest(self, inv_file, digest_algorithm, where="root"):
        """Validate the appropriate inventory digest file in path."""
        inv_digest_file = inv_file + '.' + digest_algorithm
        if not os.path.exists(inv_digest_file):
            self.log.error('E058a', where=where, path=inv_digest_file)
        else:
            self.validate_inventory_digest_match(inv_file, inv_digest_file)

    def validate_inventory_digest_match(self, inv_file, inv_digest_file):
        """Validate a given inventory digest for a given inventory file.

        On error throws exception with debugging string intended to
        be presented to a user.
        """
        if not self.check_digests:
            return
        m = re.match(r'''.*\.(\w+)$''', inv_digest_file)
        if m:
            digest_algorithm = m.group(1)
            try:
                digest_recorded = self.read_inventory_digest(inv_digest_file)
                digest_actual = file_digest(inv_file, digest_algorithm)
                if digest_actual != digest_recorded:
                    self.log.error("E060", inv_file=inv_file, actual=digest_actual, recorded=digest_recorded, inv_digest_file=inv_digest_file)
            except Exception as e:
                self.log.error("E061", description=str(e))
        else:
            self.log.error("E058b", inv_digest_file=inv_digest_file)

    def validate_object_root(self, path, version_dirs):
        """Validate object root at path.

        All expected_files must be present and no other files.
        All expected_dirs must be present and no other dirs.
        """
        expected_files = ['0=ocfl_object_1.0', 'inventory.json',
                          'inventory.json.' + self.digest_algorithm]
        for entry in os.listdir(path):
            filepath = os.path.join(path, entry)
            if os.path.isfile(filepath):
                if entry not in expected_files:
                    self.log.error('E001a', file=entry)
            elif os.path.isdir(filepath):
                if entry in version_dirs:
                    pass
                elif entry == 'extensions':
                    self.validate_extensions_dir(path)
                else:
                    self.log.error('E001b', dir=entry)
            else:
                self.log.error('E001c', entry=entry)

    def validate_extensions_dir(self, path):
        """Validate content of extensions directory inside object root at path.

        Validate the extensions directory by checking that there aren't any
        entries in the extensions directory that aren't directories themselves.
        Where there are extension directories they SHOULD be registered and
        this code relies up the registered_extensions property to list known
        extensions.
        """
        extpath = os.path.join(path, 'extensions')
        for entry in os.listdir(extpath):
            filepath = os.path.join(extpath, entry)
            if os.path.isdir(filepath):
                if entry not in self.registered_extensions:
                    self.log.warn('W013', entry=entry)
            else:
                self.log.error('E067', entry=entry)

    def validate_version_inventories(self, path, inventory, version_dirs):
        """Each version SHOULD have an inventory up to that point."""
        if len(version_dirs) == 0:
            return
        inv_digest_files = {}  # index by version_dir
        last_version = version_dirs[-1]
        for version_dir in version_dirs:
            version_path = os.path.join(path, version_dir)
            inv_file = os.path.join(version_path, 'inventory.json')
            if not os.path.exists(inv_file):
                self.log.warn('W010', where=version_dir)
            elif version_dir == last_version:
                # Don't validate in this case. Per the spec the inventory in the last version
                # MUST be identical to the copy in the object root
                root_inv_file = os.path.join(path, 'inventory.json')
                # FIXME -- how to diff efficiently?
                with open(inv_file, 'r') as ifh:
                    inv = ifh.read()
                with open(root_inv_file, 'r') as rifh:
                    root_inv = rifh.read()
                if inv != root_inv:
                    self.log.error('E064', root_inv_file=root_inv_file, inv_file=inv_file)
                else:
                    # FIXME - could just compare digest files...
                    self.validate_inventory_digest(inv_file, self.digest_algorithm, where=version_dir)
                self.inventory_digest_files[version_dir] = 'inventory.json.' + self.digest_algorithm
            else:
                # Note that inventories in prior versions may use different digest algorithms
                version_inventory, inv_validator = self.validate_inventory(inv_file, where=version_dir)
                self.validate_inventory_digest(inv_file, inv_validator.digest_algorithm, where=version_dir)
                self.inventory_digest_files[version_dir] = 'inventory.json.' + inv_validator.digest_algorithm
                # Is this inventory an appropriate prior version of the object root inventory?
                if self.root_inv_validator is not None:
                    self.root_inv_validator.validate_as_prior_version(inv_validator)

    def validate_content(self, path, inventory, version_dirs):
        """Validate file presence and content at path against inventory.

        The inventory is assumed to be valid and safe to use for construction
        of file paths etc..
        """
        files_seen = set()
        # Check on disk in each version directory
        for version_dir in version_dirs:
            version_path = os.path.join(path, version_dir)
            if not os.path.isdir(version_path):
                self.log.error('E301', version_path=version_path)
            else:
                # Check contents of version directory except content_directory
                for entry in os.listdir(version_path):
                    if ((entry == 'inventory.json')
                            or (version_dir in self.inventory_digest_files and entry == self.inventory_digest_files[version_dir])):
                        pass
                    elif entry == self.content_directory:
                        # Check content_directory
                        content_path = os.path.join(version_path, self.content_directory)
                        num_content_files_in_version = 0
                        for dirpath, dirs, files in os.walk(content_path, topdown=True):
                            if dirpath != content_path and (len(dirs) + len(files)) == 0:
                                self.log.error("E024", where=version_dir, path=dirpath)
                            for file in files:
                                obj_path = os.path.relpath(os.path.join(dirpath, file), start=path)
                                files_seen.add(obj_path)
                                num_content_files_in_version += 1
                        if num_content_files_in_version == 0:
                            self.log.warn("W003", where=version_dir)
                    elif os.path.isdir(os.path.join(version_path, entry)):
                        self.log.warn("W002", where=version_dir, entry=entry)
                    else:
                        self.log.error("E015", where=version_dir, entry=entry)
        # Check all files in root manifest
        if 'manifest' in inventory:
            for digest in inventory['manifest']:
                for filepath in inventory['manifest'][digest]:
                    if filepath not in files_seen:
                        self.log.error('E023a', where='root', content_path=filepath)
                    else:
                        if self.check_digests:
                            content_digest = file_digest(os.path.join(path, filepath), digest_type=self.digest_algorithm)
                            if content_digest != normalized_digest(digest, digest_type=self.digest_algorithm):
                                self.log.error('E092', where='root', digest=digest, content_path=filepath, content_digest=content_digest)
                        files_seen.discard(filepath)
        # Check any additional digests in root fixity block
        if 'fixity' in inventory and self.check_digests:
            for digest_algorithm in inventory['fixity']:
                for digest in inventory['fixity'][digest_algorithm]:
                    if digest != self.digest_algorithm:  # don't recheck things we check from manifest
                        for filepath in inventory['fixity'][digest_algorithm][digest]:
                            content_digest = file_digest(os.path.join(path, filepath), digest_type=digest_algorithm)
                            if content_digest != normalized_digest(digest, digest_type=digest_algorithm):
                                self.log.error('E093', where='root', digest_algorith=digest_algorithm, digest=digest, content_path=filepath, content_digest=content_digest)
        # Anything left in files_seen is not mentioned in the inventory
        if len(files_seen) > 0:
            self.log.error('E023b', where='root', extra_files=', '.join(sorted(files_seen)))

    def read_inventory_digest(self, inv_digest_file):
        """Read inventory digest from sidecar file.

        Raise exception if there is an error, else return digest.
        """
        with open(inv_digest_file, 'r') as fh:
            line = fh.readline()
            # we ignore any following lines, could raise exception
        m = re.match(r'''(\w+)\s+(\S+)\s*$''', line)
        if not m:
            raise Exception("Bad inventory digest file %s, wrong format" % (inv_digest_file))
        elif m.group(2) != 'inventory.json':
            raise Exception("Bad inventory name in inventory digest file %s" % (inv_digest_file))
        return m.group(1)
